#!/usr/bin/env python
from subprocess import Popen, PIPE, STDOUT
import shlex, sys, re, argparse, os, math
from random import randint
from threading import Thread

#Handle arguments
parser = argparse.ArgumentParser(description='Runs SPSA Training on Pachi')
parser.add_argument('-pachi', type=str, required=True, 
	help='location of pachi.exe to run')
parser.add_argument('-piter', type=str, required=True,
	help='number of iterations to use for pachi')
parser.add_argument('-fconfig', type=str, required=True,
	help='location of fuego config file')
parser.add_argument('-filedir', type=str, required=True,
	help='location of sgf files to train with')


args = parser.parse_args()

pachi_file=args.pachi
pachi_iterations=args.piter
fuego_config=args.fconfig
sgf_file_dir=args.filedir
DEBUG=False

fuego_command = "fuego --config " + fuego_config
fuego_args = shlex.split(fuego_command)

pachi_command = "" + pachi_file + " -t =" + pachi_iterations
pachi_args = shlex.split(pachi_command)

#Initial theta
theta_a = 0.0
theta_b = 0.0
theta_c = 0.0

#Variables for SPSA see spall for more info
a = 7.5			#0.1 = a/[(A+1)^alpha]
c = 0.341		#std dev centered at 0.5 with a min of 0 and max of 1 (i.e. probability of winning for a board position)
A = 1300		#10% of 13029 - # 2014 games from kgs
alpha = 0.602	#given by Spall
gamma = 0.101	#given by Spall

result = ["", ""]

#Worker thread for pachi
def pachi_script(args, input, index):
	pachi_process = Popen(args, stdout=PIPE, stdin=PIPE, stderr=STDOUT)

	pachi_output = pachi_process.communicate(input=input.encode())
	pachi_output_string = "".join(str(v) for v in pachi_output)
	pachi_output_string = pachi_output_string.replace("\\r\\n", "\n")
	op = re.search('= [a-zA-Z][0-9]{1,2}|= pass|= resign', pachi_output_string)
	result[index] = op.group(0)[2:].replace("\n", "")

k = 0
for filename in os.listdir(sgf_file_dir):
	moves_from_game = [""]
	
	#Construct a string containing all the moves from the game file
	with open(sgf_file_dir + filename) as fp:
		for i, line in enumerate(fp):
			if i>9:
				moves_from_game.append(line)

	moves_from_game="".join(moves_from_game)
	total_turns = moves_from_game.count("[")
	turn_to_train = randint(0, total_turns)
	
	color = ""
	if turn_to_train % 2 == 0:
		color = "b"
	else:
		color = "w"
	
	#choose delta
	delta_a = randint(0,1)
	delta_b = randint(0,1)
	delta_c = randint(0,1)
	
	if delta_a == 0:
		delta_a = -1
	if delta_b == 0:
		delta_b = -1
	if delta_c == 0:
		delta_c = -1
	
	#calculate c sub k
	c_subk = c/math.pow(k+1, gamma)
	
	#calculate input theta params for pachi
	plus_param_a = theta_a + c_subk*delta_a
	plus_param_b = theta_b + c_subk*delta_b
	plus_param_c = theta_c + c_subk*delta_c
	
	minus_param_a = theta_a - c_subk*delta_a
	minus_param_b = theta_b - c_subk*delta_b
	minus_param_c = theta_c - c_subk*delta_c

	
	pachi_input = "loadxmoves " + sgf_file_dir + filename + " " + str(turn_to_train) + "\nkomi 6.5\n"
	
	pachi_input_plus = pachi_input + "set-theta " + str(plus_param_a) + " " + str(plus_param_b) + " " + str(plus_param_c) + "\n"
	pachi_input_plus = pachi_input_plus + "genmove " + color + "\n"
	#pachi_input_plus = pachi_input + "evaluate-theta " + color + " " + str(plus_param_a) + " " + str(plus_param_b) + " " + str(plus_param_c) + "\n"
	t1 = Thread(target=pachi_script, kwargs={"args": pachi_args, "input": pachi_input_plus, "index": 0})
	
	pachi_input_minus = pachi_input + "set-theta " + str(minus_param_a) + " " + str(minus_param_b) + " " + str(minus_param_c) + "\n"
	pachi_input_minus = pachi_input_minus + "genmove " + color + "\n"
	#pachi_input_minus = pachi_input + "evaluate-theta " + color + " " + str(minus_param_a) + " " + str(minus_param_b) + " " + str(minus_param_c) + "\n"
	t2 = Thread(target=pachi_script, kwargs={"args": pachi_args, "input": pachi_input_minus, "index": 1})
	
	t1.start()
	t2.start()
	
	t1.join()
	t2.join()
	
	#print(result[0])
	#print(result[1])
	
	
	
	fuego_process = Popen(fuego_args, stdout=PIPE, stdin=PIPE, stderr=STDOUT)
	fuego_input = "loadsgf "+sgf_file_dir + filename +"\n"
	
	#Call undo until we get to the turn we are interested in
	i = 0
	total_turns = moves_from_game.count("[")
	while i < (total_turns - turn_to_train):
		fuego_input = fuego_input+"undo\n"
		i=i+1

	#Command(s) to run at each turn
	fuego_input = fuego_input+"play "+color+" "+result[0]+"\nfinal_score\nuct_value\n"
	fuego_input = fuego_input+"undo\nplay " + color + " " + result[1] + "\nfinal_score\nuct_value\n"
	fuego_output = fuego_process.communicate(input=fuego_input.encode())
	
	#Need to convert result from tuple to string
	fuego_output_string = "".join(str(v) for v in fuego_output)
	fuego_output_string = fuego_output_string.replace("\\r\\n", "\n")
	#print(fuego_input)
	#check_illegal = re.match('\? illegal move*', fuego_output_string)
	#if check_illegal != None:
	#	print("THE FUCK HAPPENED BRO")
	op = re.findall('= [0-9].[0-9]{1,5}', fuego_output_string)
	op[0] = 1 - float(op[0][2:])
	op[1] = 1 - float(op[1][2:])
	
	norm_diff = (op[0] - op[1])/(2*c_subk)
	g_a = norm_diff * delta_a
	g_b = norm_diff * delta_b
	g_c = norm_diff * delta_c
	
	a_subk = a/math.pow((A+k+1), alpha)
	theta_a = theta_a - a_subk * g_a
	theta_b = theta_b - a_subk * g_b
	theta_c = theta_c - a_subk * g_c
	
	k = k+1
	
	if k % 500 == 0:
		print("Multiple of 500\n")
		print(theta_a)
		print(theta_b)
		print(theta_c)
		print("\n")
		
	#if k % 100 == 0:
		
	
print("FINAL VALUES:\n")
print(theta_a)
print(theta_b)
print(theta_c)

# #Command(s) to run at that turn
# input = input+"uct_value\nuct_policy_moves\nfinal_score\nuct_value\n"

# if DEBUG:
	# print ("Constructed input string. Sending to Fuego")
	
# output = fuego_process.communicate(input=input.encode())

# #Need to convert result from tuple to string
# output_string = "".join(str(v) for v in output)
# output_string = output_string.replace("\\r\\n", "\n")

# #Extract only the lines we care about from the output
# op = re.search('= [0-9a-zA-Z.][\s\S]+', output_string) 
# result = (op.group(0))[2:]

# print ("Result from fuego is: ")
# print (result)

# #output_string = output_string[1]+output_string
# #output_string = output_string.replace("= ", "")
# #output_string = (re.split(' cannot undo 1 move.*', output_string))[0]

# #print(output_string)

